/*
 * Software für die PV-Messstation für den Betrieb in der BLE-Variante
 * Erstellt von: Jan Niklas Rother
 * Zuletzt bearbeitet am 21.02.2022
 */

#include <Wire.h>
#include "SPIFFS.h"
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <Adafruit_Sensor.h>
#include <Adafruit_BME280.h>
#include <BLEDevice.h>
#include <BLEServer.h>
#include <BLEUtils.h>
#include <BLE2902.h>

#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 64 // OLED display height, in pixels
#define SEALEVELPRESSURE_HPA (1013.25)

#define SERVICE_UUID        "4fafc201-1fb5-459e-8fcc-c5c9c331914b"
#define CHARACTERISTIC_UUID "beb5483e-36e1-4688-b7f5-ea07361b26a8"

#define TIME_TO_SLEEP 5
#define US_IN_S 1000000
#define THRESHOLD 40


Adafruit_BME280 bme; // I2C
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, -1);

BLEServer* pServer = NULL;
BLECharacteristic* pCharacteristic = NULL;

bool deviceConnected = false;
bool oldDeviceConnected = false;


class MyServerCallbacks: public BLEServerCallbacks {
    void onConnect(BLEServer* pServer) {
      deviceConnected = true;
    };

    void onDisconnect(BLEServer* pServer) {
      deviceConnected = false;
    }
};


bool touch1detected = false;

void gotTouch1(){
 touch1detected = true;
}

void setDisplay(){
  if(!display.begin(SSD1306_SWITCHCAPVCC, 0x3C))  // Address 0x3D for 128x64
  {
    Serial.println(F("SSD1306 allocation failed"));
    for(;;);
  }
  display.clearDisplay();
  display.setTextSize(1);
  display.setTextColor(WHITE);
  display.setCursor(0, 10);
  display.display(); 
}
 

void setup() {
  
  Serial.begin(115200);
  delay(2000);
  Wire.begin(23, 19);

  // Interrupts on GPIO4
  touchAttachInterrupt(T0, gotTouch1, THRESHOLD);
  
  // Setup Display
  if(!display.begin(SSD1306_SWITCHCAPVCC, 0x3C))  // Address 0x3D for 128x64
  {
    Serial.println(F("SSD1306 allocation failed"));
    for(;;);
  }
  display.clearDisplay();
  display.setTextSize(1);
  display.setTextColor(WHITE);
  display.setCursor(0, 10);
  display.display();

  // Setup BME280
  bme.begin(0x76);
  
  // Setup SPIFFS to store Data
  if(!SPIFFS.begin(true)){
    Serial.println("An Error has occurred while mounting SPIFFS");
    return;
  }
  File file = SPIFFS.open("/test.csv", FILE_WRITE);
  if(!file){
    Serial.println("There was an error opening the file for writing");
    return;
  }
  if(file.println("SAMPLE,TEMPERATURE,PRESSURE,HUMIDITY")){
    Serial.println("File was written");
  }
  else{
    Serial.println("File write failed");
  }
  file.close();
      
  // Create the BLE Device
  BLEDevice::init("ESP32");
  BLEDevice::setMTU(517);

  // Create the BLE Server
  pServer = BLEDevice::createServer();
  pServer->setCallbacks(new MyServerCallbacks());

  // Create the BLE Service
  BLEService *pService = pServer->createService(SERVICE_UUID);

  // Create a BLE Characteristic
  pCharacteristic = pService->createCharacteristic(
                      CHARACTERISTIC_UUID,
                      BLECharacteristic::PROPERTY_READ   |
                      BLECharacteristic::PROPERTY_WRITE  |
                      BLECharacteristic::PROPERTY_NOTIFY |
                      BLECharacteristic::PROPERTY_INDICATE
                    ); 


  // Create a BLE Descriptor
  pCharacteristic->addDescriptor(new BLE2902());

  // Start the service
  pService->start();

  // Start advertising
  BLEAdvertising *pAdvertising = BLEDevice::getAdvertising();
  pAdvertising->addServiceUUID(SERVICE_UUID);
  pAdvertising->setScanResponse(false);
  pAdvertising->setMinPreferred(0x0);  // set value to 0x00 to not advertise this parameter
  BLEDevice::startAdvertising();
  Serial.println("Waiting a client connection to notify...");
}

String sensordata = "";
String strSensordataBLE = "";
char sensordataBLE[40];
String temperature = "";
String pressure = "";
String humidity = "";
RTC_DATA_ATTR int counter = 1;
esp_sleep_wakeup_cause_t wakeup_reason;
  
  
void loop() {

  fetchValues();
  writeToFile();
  notifyBLE();
  displayDataOnScreen();
  counter ++;
 }

 //---------------------------------------------------

void fetchValues(){
  temperature = String(bme.readTemperature());
  pressure = String(bme.readPressure()/100.0F);
  humidity = String(bme.readHumidity());
  sensordata = String(counter) + "," + temperature + "," + pressure + "," + humidity;
  strSensordataBLE = "\n" + temperature + " °C\n" + pressure + " hPa\n" + humidity + "%";
 }


 
void displayDataOnScreen(){
  if(touch1detected){
    touch1detected = false;
    display.clearDisplay();
    display.setCursor(0, 10);
  
    display.print("Temperature: ");
    display.print(bme.readTemperature());
    display.println("C");
  
    display.print("Pressure: ");
    display.print(bme.readPressure()/100.0F);
    display.println("hPA");
  
    display.print("Humidity: ");
    display.print(bme.readHumidity());
    display.println("%");
    
    display.display();
    delay(3000);
  }
  
  display.clearDisplay();
  display.display();
 }

void writeToFile(){
  File file = SPIFFS.open("/test.csv", FILE_APPEND);

  if(file.println(sensordata))
  {
    Serial.println("File was written");
  }
    
  file.close();
}

void notifyBLE(){
  // notify changed value
  if (deviceConnected){
    strSensordataBLE.toCharArray(sensordataBLE, strSensordataBLE.length()+1);
    pCharacteristic->setValue(sensordataBLE);
    pCharacteristic->notify();
  }
      // disconnecting
    if (!deviceConnected && oldDeviceConnected) {
      delay(5000); // give the bluetooth stack the chance to get things ready
      pServer->startAdvertising(); // restart advertising
      //Serial.println("start advertising");
      oldDeviceConnected = deviceConnected;
    }
    // connecting
    if (deviceConnected && !oldDeviceConnected) {
      // do stuff here on connecting
      oldDeviceConnected = deviceConnected;
    }
  delay(5000);
}
